#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

import tornado.web
import tornado.httpserver
import tornado.ioloop

import occi
from occi.http.handler import HttpRequest, EntityHandler, CollectionHandler, DiscoveryHandler
from occi.http import HttpServer, HttpClient

class TornadoHttpServer(HttpServer):
    def __init__(self, *args, **kwargs):
        super(TornadoHttpServer, self).__init__(*args, **kwargs)

        self.application = tornado.web.Application([
            (self.base_path + r'/*/-/', TornadoRequestHandler,
                dict(handler=DiscoveryHandler(self.server, translator=self.translator))),
            (self.base_path + r'/', TornadoRequestHandler,
                dict(handler=CollectionHandler(self.server, translator=self.translator), args=[''])),
            (self.base_path + r'/(.+/)', TornadoRequestHandler,
                dict(handler=CollectionHandler(self.server, translator=self.translator))),
            (self.base_path + r'/(.+[^/])', TornadoRequestHandler,
                dict(handler=EntityHandler(self.server, translator=self.translator))),
            ])

    def run(self):
        http_server = tornado.httpserver.HTTPServer(self.application)
        http_server.listen(self.port, self.address or '')
        tornado.ioloop.IOLoop.instance().start()

class TornadoRequestHandler(tornado.web.RequestHandler):
    """Tornado RequestHandler for OCCI."""
    def __init__(self, application, request, handler=None, args=None):
        super(TornadoRequestHandler, self).__init__(application, request)
        self.handler = handler
        self.args = args

    def _handle_request(self, verb, *args):
        request = HttpRequest(
                self.request.headers.iteritems(),
                self.request.body,
                content_type=self.request.headers.get('Content-Type'),
                query_args=self.request.arguments)

        if self.args:
            args = self.args
        response = getattr(self.handler, verb)(request, *args)

#       print '%s.%s(%s): %d' % (self.handler.__class__.__name__,
#               verb, args, response.status)

        # Status code of response
        self.set_status(response.status)

        # Response Headers
        headers = {}
        for name, value in response.headers:
            values = headers.get(name)
            if values:
                values += ', ' + value
            else:
                values = value
            headers[name] = values
        for name, value in headers.iteritems():
            self.set_header(name, value)

        # Set Server header
        self.set_header('Server', occi.http.version_string)

        # Response Body
        self.write(response.body)

    def get(self, *args):
        self._handle_request('get', *args)
    def post(self, *args):
        self._handle_request('post', *args)
    def put(self, *args):
        self._handle_request('put', *args)
    def delete(self, *args):
        self._handle_request('delete', *args)


if __name__ == '__main__':
    from occi.server import OCCIServer, DummyBackend
    from occi.ext.infrastructure import *
    server = OCCIServer(backend=DummyBackend())
    server.registry.register(ComputeKind)
    server.registry.register(StorageKind)
    server.registry.register(StorageLinkKind)

    http_server = TornadoHttpServer(server, base_url='http://localhost:8000/api')
    http_server.run()



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

import re

from occi.http.utils import escape_quotes, split_quoted

class HttpHeaderError(Exception):
    pass

class HttpHeadersBase(object):
    """Base class for representing a set of HTTP headers with the same name.
    I.e. support for comma-separated multiple headers.

    >>> h = HttpHeadersBase()
    >>> h.add('text/plain')
    >>> h.add('application/xml')
    >>> h.add('text/html')
    >>> str(h)
    'text/plain, application/xml, text/html'
    >>> h.parse(str(h))
    ['text/plain', 'application/xml', 'text/html']
    >>> h.all()
    ['text/plain', 'application/xml', 'text/html']

    """
    def __init__(self):
        self._headers = []

    def add(self, header):
        self._headers.append(header)

    def parse(self, header_value):
        self._headers = [self._from_string(value.strip()) for value in split_quoted(header_value)]
        return self._headers

    def all(self):
        return self._headers

    def headers(self):
        return [self._to_string(header) for header in self._headers]

    def _from_string(self, s):
        return s

    def _to_string(self, header):
        return header

    def __str__(self):
        return ', '.join(self.headers())

class HttpAttributeHeaders(HttpHeadersBase):
    """HTTP Attribute header containing a single key-value pair per header.

    >>> h = HttpAttributeHeaders()
    >>> h.parse('foo=bar, quote="foo,bar"')
    [('foo', 'bar'), ('quote', 'foo,bar')]
    >>> h.add('escape', 'k",inject=something')
    >>> s = str(h)
    >>> s
    'foo="bar", quote="foo,bar", escape="k\\\\",inject=something"'
    >>> x = h.parse(s)
    >>> str(h) == s
    True
    >>> h.headers() == HttpHeadersBase().parse(s)
    True

    """

    def add(self, attribute, value):
        super(HttpAttributeHeaders, self).add((attribute, value))

    def _from_string(self, s):
        try:
            k, v = split_quoted(s, delimiter='=', remove_quotes=True)
        except ValueError:
            raise HttpHeaderError
        return (k, v)

    def _to_string(self, header):
        k, v = header
        if isinstance(v, int):
            value_str = str(v)
        elif isinstance(v, float):
            value_str = '%.2f' % v
        elif isinstance(v, bool):
            value_str = str(v).lower()
        else:
            value_str = '"%s"' % escape_quotes(str(v))
        return '%s=%s' % (k, value_str)

class HttpWebHeadersBase(HttpHeadersBase):
    """Base class for headers like Web Link/Category which have multiple
    key-value attributes separated by semicolon.
    """
    HEADER_VALUE_REGEXP = re.compile(r'^\s*([^;]+)\s*')

    def add(self, item, attributes=()):
        super(HttpWebHeadersBase, self).add((item, attributes))

    def _from_string(self, header_string, value_re=HEADER_VALUE_REGEXP):
        # Match header value and remove from string
        m = value_re.match(header_string)
        if not m:
            raise HttpHeaderError('%s: invalid header' % header_string)
        value = m.groups()[0]
        attr_string = header_string[m.end():]

        # Split and parse string of header attributes
        attributes = []
        for s in split_quoted(attr_string, delimiter=';'):
            if not s:
                continue
            try:
                k, v = split_quoted(s.strip(), delimiter='=', remove_quotes=True)
            except ValueError:
                raise HttpHeaderError("'%s': failed to parse header attribute" % s)
            attributes.append((k, v))

        return (value, attributes)

    def _to_string(self, header):
        item, attributes = header
        return '; '.join(
            ['%s' % item] +
            ['%s="%s"' % (name, escape_quotes(str(value))) for name, value in attributes])

class HttpLinkHeaders(HttpWebHeadersBase):
    """HTTP Web Link header.

    >>> h = HttpLinkHeaders()
    >>> h.add('/api/compute/vm01;start', [('rel', 'http://schemas.ogf.org/occi/action#start'), ('class', 'action'), ('title', 'Start')])
    >>> h.add('/api/storage/san1', [('rel', 'http://schemas.ogf.org/occi/kind#storage'), ('class', 'link'), ('title', 'Quorum Disk'), ('device', 'sda')])
    >>> s = str(h)
    >>> x = h.parse(s)
    >>> str(h) == s
    True
    >>> h.headers() == HttpHeadersBase().parse(s)
    True

    """
    HEADER_VALUE_REGEXP = re.compile(r'^\s*<([^>]+)>\s*')

    def _from_string(self, s):
        return super(HttpLinkHeaders, self)._from_string(s, value_re=self.HEADER_VALUE_REGEXP)

    def _to_string(self, header):
        uri, attributes = header
        return super(HttpLinkHeaders, self)._to_string(("<%s>" % uri, attributes))

class HttpCategoryHeaders(HttpWebHeadersBase):
    """HTTP Web Category header.

    >>> h = HttpCategoryHeaders()
    >>> h.add('compute', attributes=[('scheme', 'http://schemas.ogf.org/occi/kind#'), ('label', 'Compute Resource')])
    >>> h.add('ubuntu-9.10', [('scheme', 'http://schemas.ogf.org/occi/category#template'), ('label', 'Ubuntu Linux 9.10')])
    >>> s = str(h)
    >>> x = h.parse(s)
    >>> str(h) == s
    True
    >>> h.headers() == HttpHeadersBase().parse(s)
    True
    """
    pass

class HttpAcceptHeaders(HttpWebHeadersBase):
    """HTTP Accept header.

    >>> s = 'text/html;level=1, application/xml;q=0.9, application/xhtml+xml, text/*, image/png, */*;q=0.1'
    >>> h = HttpAcceptHeaders()
    >>> t = h.parse(s)
    >>> t == h.all()
    True
    >>> h.all()
    [('text/html', [('level', '1')]), ('application/xml', [('q', '0.9')]), ('application/xhtml+xml', []), ('text/*', []), ('image/png', []), ('*/*', [('q', '0.1')])]
    >>> [t for t, params in h.all_sorted()]
    ['text/html', 'application/xhtml+xml', 'image/png', 'application/xml', '*/*', 'text/*']
    >>> s = 'text/plain;occi=1.1;q=0.9, */*, text/html, text/*'
    >>> t = h.parse(s)
    >>> [t for t, params in h.all_sorted()]
    ['text/html', 'text/plain', 'text/*', '*/*']


    """

    def q(self, header):
        accept_type, params = header
        pref = 1.0
        if accept_type == '*/*':
            pref = -2.0
        elif accept_type.endswith('/*'):
            pref = -1.0
        try:
            for name, value in params:
                if name == 'q':
                    pref = float(value)
                    break
        except IndexError, ValueError:
            pass
        return pref

    def all_sorted(self):
        return sorted(self.all(), key=self.q, reverse=True)

if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

import optparse
import urlparse

import occi.core
import occi.http
from occi.server import OCCIServer, DummyBackend
from occi.ext.infrastructure import *
from occi.http.tornado_frontend import TornadoHttpServer

class Compute(occi.core.Resource):
    def __init__(self, kind, **kwargs):
        super(Compute, self).__init__(kind, **kwargs)
        attr_default = [
                ('occi.compute.architecture', 'x86_64'),
                ('occi.compute.speed', 2.67),
                ('occi.compute.memory', 1.0),
                ('occi.compute.state', 'inactive'),
        ]
        self.occi_set_attributes(attr_default, validate=False)

        self.occi_set_applicable_action(ComputeStartActionCategory)

    def exec_action(self, action, payload=None):
        state = self.occi_get_attribute('occi.compute.state')
        if state == 'inactive' and action.category.term == 'start':
            self.occi_set_attributes([('occi.compute.state', 'active')], validate=False)
            self.occi_set_applicable_action(ComputeStartActionCategory, applicable=False)
            self.occi_set_applicable_action(ComputeStopActionCategory, applicable=True)
        elif state == 'active' and action.category.term == 'stop':
            self.occi_set_attributes([('occi.compute.state', 'inactive')], validate=False)
            self.occi_set_applicable_action(ComputeStopActionCategory, applicable=False)
            self.occi_set_applicable_action(ComputeStartActionCategory, applicable=True)
        return None

def init_server(backend):
    ComputeKind.entity_type = Compute

    server = OCCIServer(backend=backend)
    server.registry.register(ComputeKind)
    server.registry.register(NetworkKind)
    server.registry.register(IPNetworkMixin)
    server.registry.register(StorageKind)
    server.registry.register(NetworkInterfaceKind)
    server.registry.register(IPNetworkInterfaceMixin)
    server.registry.register(StorageLinkKind)
    return server

if __name__ == "__main__":
    parser = optparse.OptionParser( description='OCCI IaaS Demo')
    parser.add_option('-B', '--base_url', dest='base_url',
            default='http://localhost:8000/',
            help='For example "http://localhost:8000/"')
    (options, args) = parser.parse_args()

    url = urlparse.urlparse(options.base_url)

    http_server = TornadoHttpServer(init_server(DummyBackend()),
            listen_address=url.hostname, listen_port=url.port,
            base_url=options.base_url)

    print "%s listen=%s port=%s" % (occi.http.version_string,
            http_server.address, http_server.port)
    print "base_url=%s base_path=%s" % (http_server.base_url + '/', http_server.base_path + '/')

    http_server.run()



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#


from occi.http.parser import Parser
from occi.http.renderer import Renderer

class JSONParser(Parser):
    """Parser for the application/json content type."""
    def parse(self, headers=None, body=None):
        raise NotImplemented('yet')

class JSONRenderer(Renderer):
    """Renderer for the application/json content type."""

    def render(self, objects):
        raise NotImplemented('yet')



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

import occi
from occi.http.parser import get_parser, register_parser, unregister_parser
from occi.http.renderer import get_renderer, register_renderer, unregister_renderer
from occi.http.dataobject import URLTranslator

# OCCI Http Rendering version
__http_version__ = '1.1'

# HTTP version string for use in Server/Client headers
version_string = 'occi-py/%s OCCI/%s' % (occi.__version__, __http_version__)

class HttpServer(object):
    def __init__(self, occi_server,
            listen_address=None, listen_port=None,
            base_url=None):
        self.server = occi_server
        self.address = listen_address
        self.port = listen_port or 80
        self.translator = URLTranslator(base_url or '')
        self.base_url = self.translator.base_url
        self.base_path = self.translator.base_path

    def run(self):
        raise NotImplementedError

class HttpClient(object):
    def __init__(self, occi_client,
            base_url=None):
        self.client = occi_client
        self.base_url = base_url




#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

import urlparse

from occi.core import Category, Kind, Mixin, Resource, Link, Action, IDTranslator

class DataObject(object):
    """A data object transferred using the OCCI protocol.

    A data object cat represent a resource instance, an action invocation,
    filter parameters, etc. It is up to the handler of the particular request/response
    to interpret the contents of a `DataObject`.
    """

    class DataObjectError(Exception):
        pass
    class Invalid(DataObjectError):
        pass

    def __init__(self, categories=None, attributes=None, links=None,
            location=None, translator=None):
        self.categories = categories or []
        self.links = links or []
        self.attributes = attributes or []
        self.location = location

        self.translator = translator or URLTranslator('')
        self.parse_flags = {}
        self.render_flags = {}

    def load_from_entity(self, entity):
        """Load `DataObject` with the contents of the specified Entity instance.

        >>> from occi.ext.infrastructure import *
        >>> compute = ComputeKind.entity_type(ComputeKind)
        >>> compute.id = 'compute/123'
        >>> compute.occi_set_attributes([('occi.compute.speed', 7.0/3)], validate=False)
        >>> compute.occi_set_applicable_action(ComputeStartActionCategory)
        >>> storage = StorageKind.entity_type(StorageKind)
        >>> storage.id = 'storage/234'
        >>> storage.occi_set_attributes([('title', 'My Disk')], validate=False)
        >>> link = StorageLinkKind.entity_type(StorageLinkKind)
        >>> link.id = 'link/storage/345'
        >>> link.target = storage
        >>> link.occi_set_attributes([('occi.storagelink.deviceid', 'ide:0:1')], validate=False)
        >>> compute.links.append(link)
        >>> d = DataObject(translator=URLTranslator('/api/'))
        >>> d.load_from_entity(compute)
        >>> d.location
        '/api/compute/123'
        >>> d.categories
        [Kind('compute', 'http://schemas.ogf.org/occi/infrastructure#')]
        >>> d.attributes
        [('occi.compute.speed', 2.3333333333333335)]
        >>> [(l.target_location, l.target_categories, l.target_title) for l in d.links]
        [('/api/storage/234', [Kind('storage', 'http://schemas.ogf.org/occi/infrastructure#')], 'My Disk'), ('/api/compute/123?action=start', [Category('start', 'http://schemas.ogf.org/occi/infrastructure/compute/action#')], 'Start Compute Resource')]
        >>> [(l.link_location, l.link_categories, l.link_attributes) for l in d.links]
        [('/api/link/storage/345', [Kind('storagelink', 'http://schemas.ogf.org/occi/infrastructure#')], [('occi.storagelink.deviceid', 'ide:0:1')]), (None, [], [])]

        """
        # Set location translator for Entity instance
        entity.occi_set_translator(self.translator)

        # Get Entity Kind, Mixins, Attributes and ID
        self.categories = entity.occi_list_categories()
        self.attributes = entity.occi_get_attributes(convert=True)
        self.location = self.translator.from_native(entity.id)

        # Links
        if isinstance(entity, Resource):
            for link in entity.links:
                l = LinkRepr(
                        target_location=self.translator.from_native(link.target.id),
                        target_categories=link.target.occi_list_categories(),
                        target_title=link.target.occi_get_attribute('title'),
                        link_location=self.translator.from_native(link.id))
                link_attributes = link.occi_get_attributes(convert=True,
                        exclude=('source', 'target'))
                if link_attributes:
                    l.link_categories = link.occi_list_categories()
                    l.link_attributes = link_attributes

                self.links.append(l)

        # Actions
        for action in entity.occi_list_applicable_actions():
            l = LinkRepr(
                    target_location='%s?action=%s' % (
                        self.translator.from_native(entity.id, path_only=True),
                        action.term),
                    target_categories=[action],
                    target_title=action.title)
            self.links.append(l)

    def save_to_entity(self, entity=None, category_registry=None,
            validate_attr=True, save_links=False):
        """Save the `DataObject` contents into an Entity instance.

        >>> from occi.ext.infrastructure import *
        >>> d = DataObject(translator=URLTranslator('/api'))
        >>> d.location = '/api/compute/123'
        >>> d.categories = [ComputeKind]
        >>> d.attributes = [('occi.compute.speed', '2.33')]
        >>> l = LinkRepr(target_location='/api/storage/234', target_categories=[StorageKind])
        >>> l.link_location = '/api/link/storage/345'
        >>> l.link_categories = [StorageLinkKind]
        >>> l.link_attributes = [('occi.storagelink.deviceid', 'ide:0:1')]
        >>> d.links.append(l)
        >>> entity = d.save_to_entity(save_links=True)
        >>> entity.occi_list_categories()
        [Kind('compute', 'http://schemas.ogf.org/occi/infrastructure#')]
        >>> round(entity.occi_get_attribute('occi.compute.speed')*1000)
        2330.0
        >>> entity.links[0].occi_list_categories()
        [Kind('storagelink', 'http://schemas.ogf.org/occi/infrastructure#')]
        >>> entity.links[0].occi_get_attributes()
        [('source', 'compute/123'), ('target', 'storage/234'), ('occi.storagelink.deviceid', 'ide:0:1')]
        >>> entity.links[0].target.occi_list_categories()
        [Kind('storage', 'http://schemas.ogf.org/occi/infrastructure#')]
        >>> entity.links[0].target.occi_get_attributes()
        []
        """

        # Resolve categories
        try:
            kind, mixins = self._resolve_categories(self.categories,
                    category_registry=category_registry)
        except Category.DoesNotExist as e:
            raise self.Invalid(e)

        # Create new entity if not specified already
        if not entity:
            # Kind instance required
            if not kind:
                raise self.Invalid('Kind not specified, cannot create Entity')
            entity = kind.entity_type(kind, mixins=mixins)
        else:
            if kind and str(kind) != str(entity.occi_get_kind()):
                raise self.Invalid('Cannot change Kind of existing Entity')
            [entity.occi_add_mixin(mixin) for mixin in mixins]

        # Set location translator for Entity instance
        entity.occi_set_translator(self.translator)

        # Load attributes
        entity.occi_set_attributes(self.attributes, validate=validate_attr)

        # Load Link relations
        if save_links and self.links:
            if not isinstance(entity, Resource):
                raise self.Invalid('Links only applicable to the Resource type')
            for link_repr in self.links:
                # Initialise target Resource
                t_kind, t_mixins = self._resolve_categories(
                        link_repr.target_categories, category_registry)
                target = t_kind.entity_type(t_kind, mixins=t_mixins)
                target.occi_set_translator(self.translator)
                if not isinstance(target, Resource):
                    raise self.Invalid('Link target must be a Resource type')
                target.id = self.translator.to_native(link_repr.target_location)

                # Initialise Link instance
                try:
                    l_kind, l_mixins = self._resolve_categories(
                            link_repr.link_categories, category_registry)
                except self.Invalid:
                    l_kind = LinkKind
                    l_mixins = []
                link = l_kind.entity_type(l_kind, mixins=l_mixins)
                link.occi_set_translator(self.translator)
                if not isinstance(link, Link):
                    raise self.Invalid('Relation must be a Link type')
                link.id = self.translator.to_native(link_repr.link_location)
                link.target = target
                default_attr = [
                        ('source', self.location),
                        ('target', link_repr.target_location)
                ]
                link.occi_set_attributes(
                        default_attr + link_repr.link_attributes,
                        validate=validate_attr)

                # Add Link instance to the Resource's list of links
                entity.links.append(link)

        return entity

    def _resolve_categories(self, categories, category_registry=None):
        """Extract `Kind` and `Mixin`s from a list of Categories. If specified,
        resolve each `Category` through the `CategoryRegistry`.
        """
        kind = None
        mixins = []

        # Resolve Categories and extract Kind
        for category in categories:
            if category_registry:
                category = category_registry.lookup_id(str(category))
            if isinstance(category, Kind):
                if kind is not None:
                    raise self.Invalid('%s: Only one Kind allowed to define a resource' % category)
                kind = category
            elif isinstance(category, Mixin):
                mixins.append(category)
            else:
                raise self.Invalid('%s: Is neither a Kind nor a Mixin' % category)

        return kind, mixins

    def save_as_action(self, category_registry=None):
        """Save the `DataObject` contents as an Action instance.

        >>> from occi.ext.infrastructure import *
        >>> d = DataObject()
        >>> d.categories = [ComputeStartActionCategory]
        >>> d.attributes = [('method', 'acpioff')]
        >>> action = d.save_as_action()
        >>> action.category
        Category('start', 'http://schemas.ogf.org/occi/infrastructure/compute/action#')
        >>> action.parameters
        [('method', 'acpioff')]
        >>> d.attributes.append(('foo', 'bar'))
        >>> d.save_as_action()
        Traceback (most recent call last):
        Invalid: "foo": Unknown parameter
        """

        # Resolve category
        if len(self.categories) != 1:
            raise self.Invalid('Specify a single Category to identify an Action')
        if category_registry:
            try:
                category = category_registry.lookup_id(self.categories[0])
            except Category.DoesNotExist as e:
                raise self.Invalid(e)
        else:
            category = self.categories[0]

        # Create new Action instance
        try:
            action = Action(category, self.attributes)
        except Action.ActionError as e:
            raise self.Invalid(e)

        return action

class LinkRepr(object):
    def __init__(self,
            target_location=None, target_title=None, target_categories=None,
            link_location=None, link_categories=None, link_attributes=None):
        self.target_location = target_location
        self.target_title = target_title
        self.target_categories = target_categories or []
        self.link_location = link_location
        self.link_categories = link_categories or []
        self.link_attributes = link_attributes or []

class URLTranslator(IDTranslator):
    """Translates between Entity ID and Location URL"""
    def __init__(self, base_url):
        t = urlparse.urlparse(base_url.rstrip('/'))
        self.base_url = t.geturl()
        self.base_path = t.path.rstrip('/')

    def from_native(self, entity_id, path_only=False):
        if path_only:
            return '%s/%s' % (self.base_path, entity_id)
        return '%s/%s' % (self.base_url, entity_id)

    def to_native(self, location):
        i = 0
        if location.startswith(self.base_url):
            i = len(self.base_url)
        elif location.startswith(self.base_path):
            i = len(self.base_path)
        return location[i:].lstrip('/')

if __name__ == "__main__":
    import doctest
    doctest.testmod()



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

try:
    from ordereddict import OrderedDict
except ImportError:
    OrderedDict = dict

class Attribute(object):
    class Invalid(Exception):
        def __init__(self, name, value):
            self.name = name
            self.value = value

        def __str__(self):
            return "%s='%s': invalid attribute value" % (self.name, self.value)

    def __init__(self, name, required=True, mutable=False):
        self.name = name
        self.required = required
        self.mutable = mutable

    def to_native(self, s, **kwargs):
        return s

    def from_native(self, v, **kwargs):
        return v

    def __repr__(self):
        return "%s('%s', required=%s, mutable=%s)" % (self.__class__.__name__, self.name, self.required, self.mutable)

class IntAttribute(Attribute):
    def to_native(self, s, **kwargs):
        try:
            return int(s)
        except ValueError:
            raise self.Invalid(self.name, s)

class FloatAttribute(Attribute):
    def to_native(self, s, **kwargs):
        try:
            return float(s)
        except ValueError:
            raise self.Invalid(self.name, s)

class BoolAttribute(Attribute):
    def to_native(self, s, **kwargs):
        if s == 'true':
            return True
        elif s == 'false':
            return False
        else:
            raise self.Invalid(self.name, s)

class IDAttribute(Attribute):
    def to_native(self, s, translator=None, **kwargs):
        return translator.to_native(s)

    def from_native(self, s, translator=None, **kwargs):
        return translator.from_native(s)

class Category(object):
    """The OCCI Category type."""

    class CategoryError(Exception):
        pass
    class Invalid(CategoryError):
        pass
    class DoesNotExist(CategoryError):
        pass

    def __init__(self, term, scheme, title=None, related=None, attributes=None):
        self.term = term
        self.scheme = scheme
        self.title = title
        self.related = related
        self.attributes = []
        self.unique_attributes = []

        # Attributes
        if related:
            self.attributes.extend(related.attributes)
        if attributes:
            self.attributes.extend(attributes)
            self.unique_attributes = attributes

    def __repr__(self):
        return "%s('%s', '%s')" % (self.__class__.__name__, self.term, self.scheme)

    def __str__(self):
        return self.scheme + self.term

    def __cmp__(self, other):
        return cmp(str(self), str(other))

    def is_related(self, category):
        current = self
        while current:
            if category == current:
                return True
            current = current.related
        return False

class CategoryRegistry(object):
    """Registry of all Category/Kind/Mixin instances currently known to the
    OCCI server or client.

    >>> reg = CategoryRegistry()
    >>> from occi.core import Category, ExtCategory, Kind, Mixin
    >>> from occi.ext.infrastructure import *
    >>> reg.register(ComputeKind)
    >>> reg.register(StorageKind)
    >>> reg.register(StorageLinkKind)
    >>> fooKind = Kind('foo', 'http://#', related=ResourceKind, location='compute/')
    >>> reg.register(fooKind)
    Traceback (most recent call last):
    Invalid: compute/: location path already defined
    >>> reg.lookup_id(ComputeKind)
    Kind('compute', 'http://schemas.ogf.org/occi/infrastructure#')
    >>> reg.lookup_location('storage/')
    Kind('storage', 'http://schemas.ogf.org/occi/infrastructure#')
    >>> reg.unregister(StorageKind)
    >>> reg.unregister(ComputeKind)
    >>> reg.unregister(EntityKind) ; reg.unregister(ResourceKind) ; reg.unregister(LinkKind)
    >>> reg.all()
    [Kind('storagelink', 'http://schemas.ogf.org/occi/infrastructure#')]

    """

    def __init__(self):
        self._categories = OrderedDict()
        self._locations = {}

        # Always register OCCI Core types
        self.register(EntityKind)
        self.register(ResourceKind)
        self.register(LinkKind)

    def register(self, category):
        """Register a new Category/Kind/Mixin."""
        s = str(category)
        if s in self._categories:
            raise Category.Invalid('%s: Category already registered' % s)

        # Location
        if hasattr(category, 'location') and category.location:
            if category.location in self._locations:
                raise Category.Invalid('%s: location path already defined' % category.location)
            self._locations[category.location] = category

        # Register category
        self._categories[s] = category

        # Register actions
        if hasattr(category, 'actions'):
            for action in category.actions:
                if hasattr(action, 'actions'):
                    raise Category.Invalid(
                            '%s: Only the base Category type allowed to identify Actions' % action)
                self.register(action)

    def unregister(self, category):
        """Unregister a previously registered Category/Kind/Mixin."""
        try:
            category = self._categories[str(category)]
        except KeyError:
            raise Category.Invalid("%s: Category not registered" % category)

        # Unregister category
        del self._categories[str(category)]

        # Remove location entry
        if hasattr(category, 'location') and category.location:
            self._locations.pop(category.location, None)

        # Remove additional action categories
        if hasattr(category, 'actions'):
            for action in category.actions:
                self.unregister(action)

    def lookup_id(self, identifier):
        try:
            return self._categories[str(identifier)]
        except KeyError:
            raise Category.DoesNotExist('"%s": Category does not exist' % identifier)

    def lookup_location(self, path):
        loc = path.lstrip('/')
        return self._locations.get(loc)

    def all(self):
        return self._categories.values()

class ExtCategory(Category):
    def __init__(self, term, scheme, actions=None, location=None, **kwargs):
        super(ExtCategory, self).__init__(term, scheme, **kwargs)
        self._location = None
        self.actions = actions or ()
        self.location = location

    def get_location(self):
        return self._location
    def set_location(self, loc):
        if not loc:
            loc = None
        elif loc[-1] != '/':
            raise Category.Invalid("%s: invalid location path, must end with '/'" % loc)
        else:
            loc = loc.lstrip('/')
        if not loc:
            loc = None
        self._location = loc
    location = property(get_location, set_location)

class Kind(ExtCategory):
    """The OCCI Kind type.

    A Kind instance uniquely identifies an Entity sub-type.
    """
    def __init__(self, term, scheme, entity_type=None, **kwargs):
        super(Kind, self).__init__(term, scheme, **kwargs)
        self.entity_type = entity_type or Entity

        if self.related and not isinstance(self.related, Kind):
            raise Category.Invalid("Kind instance can only be related to other Kind instances")

class Mixin(ExtCategory):
    """The OCCI Mixin type.

    A Mixin instance adds additional capabilities (attributes and actions) to
    an existing resource instance. A 'resource instance' is an instance of a
    sub-type of Entity.
    """
    def __init__(self, term, scheme, userdefined=None, **kwargs):
        super(Mixin, self).__init__(term, scheme, **kwargs)
        self.userdefined = userdefined or False

        if self.related and not isinstance(self.related, Mixin):
            raise Category.Invalid("Mixin instance can only be related to other Mixin instances")

class Action(object):
    """The OCCI Action type.

    An instance of Action represents a operation (on a resource instance) to be
    invoked. An Action instance consist of its defining Category and a set of
    attribute (parameter) values. The available parameters are defined by the Category.

    >>> startAction = Category('start', 'http://example.com/occi/foo/action#', attributes=[Attribute('example.com.bar', required=True)])
    >>> action = Action(startAction, parameters=[('example.com.bar', 'foobar')])
    >>> action.parameters
    [('example.com.bar', 'foobar')]
    >>> action = Action(startAction, parameters=[('example.com.unknown', 'blah'), ('example.com.bar', 'foobar')])
    Traceback (most recent call last):
    UnknownParameter: "example.com.unknown": Unknown parameter
    >>> action = Action(startAction, parameters=[('example.com.unknown', 'blah')])
    Traceback (most recent call last):
    RequiredParameter: "example.com.bar": Required parameter
    """

    class ActionError(Exception):
        def __init__(self, item=None, message=None):
            self.item = item
            self.message = message
        def __str__(self):
            s = ''
            if self.item:
                s += '"%s": ' % self.item
            if hasattr(self, '_name'):
                s += self._name
            else:
                s += self.__class__.__name__
            if self.message:
                s += ': %s' % self.message
            return s
    class UnknownParameter(ActionError):
        _name = 'Unknown parameter'
    class RequiredParameter(ActionError):
        _name = 'Required parameter'

    def __init__(self, category, parameters=None):
        """Setup Action instance given its defining Category and a list of
        parameter values.

        :param category: Category instance defining the Action.
        :keyword parameters: List of parameter name-value pairs.
        """
        self._occi_category = category
        self._occi_parameters = {}

        # Load supplied parameters into a dictionary
        param_dict = {}
        for param, value in parameters or ():
            if param in param_dict:
                raise self.DuplicateAttribute(param)
            param_dict[param] = value

        # Setup the parameters for the Action instance
        for attribute in self._occi_category.attributes:
            try:
                value = param_dict[attribute.name]
                del param_dict[attribute.name]
            except KeyError:
                pass
            else:
                self._occi_parameters[attribute.name] = attribute.to_native(value)

            # Check required parameters
            if attribute.required and attribute.name not in self._occi_parameters:
                raise self.RequiredParameter(attribute.name)

        # Verify all supplied attributes have been handled
        if param_dict:
            raise self.UnknownParameter(param_dict.keys()[0])

    def _get_occi_category(self):
        return self._occi_category
    category = property(_get_occi_category)

    def _get_occi_parameters(self):
        params = []
        for attribute in self._occi_category.attributes:
            try:
                params.append((attribute.name, self._occi_parameters[attribute.name]))
            except KeyError:
                pass
        return params
    parameters = property(_get_occi_parameters)

class IDTranslator(object):
    """Translate Entity ID between native (internal) and external
    representation.

    Extend this class to implement another translation mechanism. Simply
    override the to_native() and from_native methods.

    The OCCI Http Rendering provides an URL translator suitable for mapping
    URLs to Entity IDs. See occi.http.URLTranslator.
    """
    def to_native(self, ext, **kwargs):
        return str(ext)
    def from_native(self, entity_id, **kwargs):
        return str(entity_id)

class Entity(object):
    """The OCCI Entity (abstract) type.

    The abstract Entity type is inherited by the Resource and Link types. The
    Entity type and any sub-type thereof MUST have its own unique identifying
    Kind instance.

    A 'resource instance' is an instance of a sub-type of Entity.
    """

    class EntityError(Exception):
        def __init__(self, item=None, message=None):
            self.item = item
            self.message = message
        def __str__(self):
            s = ''
            if self.item:
                s += '"%s": ' % self.item
            if hasattr(self, '_name'):
                s += self._name
            else:
                s += self.__class__.__name__
            if self.message:
                s += ': %s' % self.message
            return s
    class DoesNotExist(EntityError):
        _name = 'Resource instance does not exist'
    class UnknownCategory(EntityError):
        _name = 'Unknown Category'
    class InvalidCategory(EntityError):
        _name = 'Invalid Category'
    class UnknownAttribute(EntityError):
        _name = 'Unknown attribute'
    class DuplicateAttribute(EntityError):
        _name = 'Duplicate attribute'
    class ImmutableAttribute(EntityError):
        _name = 'Immutable attribute'
    class RequiredAttribute(EntityError):
        _name = 'Required attribute'

    def __init__(self, kind, mixins=[]):
        self.id = None
        self._occi_kind = None
        self._occi_mixins = {}
        self._occi_attributes = {}
        self._occi_actions_available = {}
        self._occi_actions_applicable = {}
        self._occi_translator = IDTranslator()

        # Set the Kind of this resource instance
        if not kind or not isinstance(kind, Kind) or not kind.is_related(EntityKind):
            raise self.InvalidCategory(kind, 'not a valid Kind instance')
        self._occi_kind = kind
        self._add_actions_available(kind.actions)

        # Add additional Mixins
        for mixin in mixins:
            self.occi_add_mixin(mixin)

    def _add_actions_available(self, actions=[]):
        for category in actions:
            cat_id = str(category)
            self._occi_actions_available[cat_id] = category

    def _remove_actions_available(self, actions=[]):
        for category in actions:
            cat_id = str(category)
            self._occi_actions_available.pop(cat_id, None)
            self._occi_actions_applicable.pop(cat_id, None)

    def occi_get_kind(self):
        return self._occi_kind

    def occi_add_mixin(self, mixin):
        cat_id = str(mixin)

        # Must be a Mixin type
        if not isinstance(mixin, Mixin):
            raise self.InvalidCategory(mixin, 'not a Mixin instance')

        # Save mixin
        self._occi_mixins[cat_id] = mixin
        self._add_actions_available(mixin.actions)

    def occi_remove_mixin(self, mixin):
        try:
            mixin = self._occi_mixins.pop(str(mixin))
            self._remove_actions_available(mixin.actions)
        except KeyError:
            raise self.UnknownCategory(mixin, 'not found')

    def occi_list_categories(self):
        return [self._occi_kind] + self._occi_mixins.values()

    def occi_set_translator(self, translator):
        self._occi_translator = translator

    def occi_get_attribute(self, name):
        """Get single OCCI attribute value."""
        return self._occi_attributes.get(name)

    def occi_get_attributes(self, convert=False, exclude=()):
        """Get list of OCCI attribute key-value pairs.

        :keyword convert: If True convert from OCCI native format to external
            representation.
        :keyword exclude: A list of attribute names to exclude.
        """
        attr_list = []
        for category in self.occi_list_categories():
            for attribute in category.attributes:
                if attribute.name in exclude:
                    continue
                try:
                    value = self._occi_attributes[attribute.name]
                except KeyError:
                    pass
                else:
                    if convert:
                        value = attribute.from_native(value,
                                translator=self._occi_translator)
                    attr_list.append((attribute.name, value))
        return attr_list

    def occi_set_attributes(self, attr_list, validate=True):
        """Set the values of the OCCI attributes defined for this resource
        instance.

        :param attr_list: List of key-value tuples
        :param validate: Boolean whether to validate the attribute set

        >>> fooKind = Kind('foo', 'http://example.com/occi#', title='Foo', related=ResourceKind, attributes=[Attribute('com.example.bar', required=True, mutable=True)])
        >>> entity = Entity(fooKind)
        >>> attrs = [('summary', 'blah blah')]
        >>> entity.occi_set_attributes(attrs, validate=True)
        Traceback (most recent call last):
            File "core.py", line 362, in occi_set_attributes
                raise self.RequiredAttribute(attribute.name)
        RequiredAttribute: "com.example.bar": Required attribute
        >>> attrs += [('title', 'A "tiny" resource instance')]
        >>> attrs += [('com.example.bar', 'Bar')]
        >>> entity.occi_set_attributes(attrs, validate=True)
        >>> entity.occi_get_attributes(convert=True)
        [('title', 'A "tiny" resource instance'), ('summary', 'blah blah'), ('com.example.bar', 'Bar')]
        >>> attrs += [('summary', 'duplicate')]
        >>> entity.occi_set_attributes(attrs, validate=True)
        Traceback (most recent call last):
            File "core.py", line 256, in occi_set_attributes
                raise self.DuplicateAttribute(attr)
        DuplicateAttribute: "summary": Duplicate attribute

        """
        # Load supplied attributes into a dictionary
        attr_dict = {}
        for attr, value in attr_list:
            if attr in attr_dict:
                raise self.DuplicateAttribute(attr)
            attr_dict[attr] = value

        # Add attributes to the Entity instance
        for category in self.occi_list_categories():
            for attribute in category.attributes:
                try:
                    value = attr_dict[attribute.name]
                    del attr_dict[attribute.name]
                except KeyError:
                    pass
                else:
                    # Attribute mutable if:
                    #  - attribute.mutable == True
                    #  - attribute.required == True and attribute.mutable == False and attribute not yet specified (write once)
                    if not validate or attribute.mutable or (
                            attribute.required and attribute.name not in self._occi_attributes):
                        # Convert and save new attibute value
                        self._occi_attributes[attribute.name] = attribute.to_native(value,
                                translator=self._occi_translator)
                    else:
                        raise self.ImmutableAttribute(attribute.name)

                # Check required attribute
                if validate and attribute.required and attribute.name not in self._occi_attributes:
                    raise self.RequiredAttribute(attribute.name)

        # Verify all supplied attributes have been handled
        if attr_dict:
            raise self.UnknownAttribute(attr_dict.keys()[0])

    def occi_list_actions(self):
        """Return a list of Category instances which define the Actions
        _available_ to this resource instance.
        """
        return self._occi_actions_available.values()

    def occi_list_applicable_actions(self):
        """Return a list of Category instances which define the Actions
        currently _applicable_ to this resource instance.
        """
        return [self._occi_actions_available[cat_id] for cat_id in self._occi_actions_applicable.keys()]

    def occi_is_applicable_action(self, action_category):
        """Return whether the given Category instance correspond to a currently
        applicable Action.
        """
        cat_id = str(action_category)
        return cat_id in self._occi_actions_applicable and cat_id in self._occi_actions_available

    def occi_set_applicable_action(self, action_category, applicable=True):
        """Set 'applicable' state of an action. By default all actions defined
        for a resource instance are non-applicable.

        :param action_category: The Category instance defining the Action.
        :keyword applicable: Boolean, whether action is currently applicable or not.

        >>> startAction = Category('start', 'http://example.com/occi/foo/action#')
        >>> fooKind = Kind('foo', 'http://example.com/occi#', title='Foo', related=ResourceKind, attributes=[Attribute('com.example.bar', required=True, mutable=True)], actions=[startAction])
        >>> entity = Entity(fooKind)
        >>> entity.occi_list_actions()
        [Category('start', 'http://example.com/occi/foo/action#')]
        >>> entity.occi_list_applicable_actions()
        []
        >>> entity.occi_is_applicable_action(startAction)
        False
        >>> entity.occi_set_applicable_action(startAction)
        >>> entity.occi_is_applicable_action(startAction)
        True
        >>> entity.occi_list_applicable_actions()
        [Category('start', 'http://example.com/occi/foo/action#')]
        >>> stopAction = Category('stop', 'http://example.com/occi/foo/action#')
        >>> entity.occi_is_applicable_action(stopAction)
        False
        >>> entity.occi_set_applicable_action(stopAction)
        Traceback (most recent call last):
        UnknownCategory: "http://example.com/occi/foo/action#stop": Unknown Category: Action not defined for this resource instance

        """
        cat_id = str(action_category)
        if cat_id not in self._occi_actions_available:
            raise self.UnknownCategory(cat_id, 'Action not defined for this resource instance')
        if applicable:
            self._occi_actions_applicable[cat_id] = True
        else:
            self._occi_actions_applicable.pop(cat_id, None)

class Resource(Entity):
    def __init__(self, kind, links=None, **kwargs):
        super(Resource, self).__init__(kind, **kwargs)
        self.links = links or []

class Link(Entity):
    def __init__(self, kind, target=None, **kwargs):
        super(Link, self).__init__(kind, **kwargs)
        self.target = target

EntityKind = Kind('entity', 'http://schemas.ogf.org/occi/core#',
        title='Entity type',
        entity_type=Entity,
        attributes=(
            Attribute('title', required=False, mutable=True),
        ),
)

ResourceKind = Kind('resource', 'http://schemas.ogf.org/occi/core#',
        related=EntityKind,
        title='Resource type',
        entity_type=Resource,
        attributes=(
            Attribute('summary', required=False, mutable=True),
        ),
)

LinkKind = Kind('link', 'http://schemas.ogf.org/occi/core#',
        related=EntityKind,
        title='Link type',
        entity_type=Link,
        attributes=(
            IDAttribute('source', required=True, mutable=False),
            IDAttribute('target', required=True, mutable=True),
        ),
)

ActionCategory = Category('action', 'http://schemas.ogf.org/occi/core#',
        title='Action')


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

def escape_quotes(s, quotechar='"', escapechar='\\'):
    """Escape quote character and also escape the escape character itself.

    >>> escape_quotes('This is a "little" VM')
    'This is a \\\\"little\\\\" VM'
    """
    buf = ''
    for c in s:
        if c == quotechar or c == escapechar:
            buf += escapechar
        buf += c
    return buf

def split_quoted(s, delimiter=',', quotechar='"', escapechar='\\', remove_quotes=False):
    """Split string on delimiter if and only if delimiter is not within a quote
    nor escaped by the escape character. An escaped quote character will not
    affect the quote state during parsing.

    >>> split_quoted(r'name="foo,bar",size=40, key="value",')
    ['name="foo,bar"', 'size=40', ' key="value"']
    >>> split_quoted(r'name="foo,bar",size=40, key="value",', remove_quotes=True)
    ['name=foo,bar', 'size=40', ' key=value']
    >>> split_quoted(r'name="foo,bar\\",size=40, key=", value=bar\\,foo, items=3', remove_quotes=True)
    ['name=foo,bar",size=40, key=', ' value=bar,foo', ' items=3']
    """
    l = []
    quote = False
    escape = False
    buf = None
    for c in s:
        #print "e=%s q=%s\t%s" % (escape, quote, buf)
        if c == escapechar:
            escape = not escape
            if escape and remove_quotes:
                if buf is None:
                    buf = ''
                continue
        elif c == quotechar and not escape:
            quote = not quote
            if remove_quotes:
                if buf is None:
                    buf = ''
                continue
        elif c == delimiter and not escape and not quote:
            l.append(buf)
            buf = None
            continue
        elif escape:
            escape = False

        if buf is None:
            buf = ''
        buf += c
    if buf is not None:
        l.append(buf)
    return l


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

from occi.core import Category, Entity, Mixin
from occi.server import ServerBackend
from occi.http import get_parser, get_renderer
from occi.http.header import HttpHeaderError
from occi.http.parser import ParserError
from occi.http.renderer import RendererError
from occi.http.dataobject import DataObject

class HttpRequest(object):
    def __init__(self, headers, body, content_type=None,
            user=None, query_args=None):
        self.headers = headers
        self.body = body
        self.content_type = content_type
        self.user = user
        self.query_args = query_args or {}

class HttpResponse(object):
    def __init__(self, headers=None, body=None, status=None):
        self.status = status or 200
        self.headers = headers or []
        self.body = body or ''

class HttpRequestError(Exception):
    """Exception wrapper for returning a proper HTTP response if an error
    occurs."""
    def __init__(self, response=None):
        assert(isinstance(response, HttpResponse))
        self.response = response
    def __str__(self):
        return repr(self.response)

class HttpResponseCode(object):
    def http_response(self, status, msg=None):
        headers = [('Content-Type', 'text/plain; charset=utf-8')]
        msg = msg or ''
        return HttpResponse(status=status, headers=headers, body=str(msg))

    def ALL_OK(self, msg='OK'):
        return self.http_response(200, msg)
    def CREATED(self, msg='Created'):
        return self.http_response(201, msg)
    def ACCEPTED(self, msg='Accepted'):
        return self.http_response(202, msg)
    def DELETED(self, msg=''):
        return self.http_response(204, msg)
    def BAD_REQUEST(self, msg='Bad request'):
        return self.http_response(400, msg)
    def FORBIDDEN(self, msg='Forbidden'):
        return self.http_response(401, msg)
    def NOT_FOUND(self, msg='Not Found'):
        return self.http_response(404, msg)
    def CONFLICT(self, msg='Conflict/Duplicate'):
        return self.http_response(401, msg)
    def NOT_HERE(self, msg='Gone'):
        return self.http_response(410, msg)
    def SERVER_ERROR(self, msg='Internal Server Error'):
        return self.http_response(500, msg)
    def NOT_IMPLEMENTED(self, msg='Not Implemented'):
        return self.http_response(501, msg)
    def THROTTLED(self, msg='Throttled'):
        return self.http_response(503, msg)
hrc = HttpResponseCode()

class HandlerBase(object):
    """HTTP handler base class."""

    def __init__(self, server, translator=None):
        self.server = server            # OCCIServer
        self.translator = translator    # URLTranslator

    def _request_init(self, request):
        """Parse request and initialize response renderer."""
        try:
            parser = get_parser(request.content_type,
                    translator=self.translator)
            parser.parse(request.headers, request.body)
        except (ParserError, HttpHeaderError) as e:
            raise HttpRequestError(hrc.BAD_REQUEST(e))

        # Get renderer
        try:
            renderer = get_renderer(parser.accept_types)
        except RendererError as e:
            raise HttpRequestError(hrc.BAD_REQUEST(e))

        return parser, renderer

    def _get_entity(self, entity_id, user=None):
        """Load entity object from backend."""
        try:
            return self.server.backend.get_entity(entity_id, user=user)
        except Entity.DoesNotExist as e:
            raise HttpRequestError(hrc.NOT_FOUND(e))
        except ServerBackend.InvalidOperation as e:
            raise HttpRequestError(hrc.BAD_REQUEST(e))
        except ServerBackend.ServerBackendError as e:
            print e
            raise HttpRequestError(hrc.SERVER_ERROR())

    def _filter_entities(self, categories=None, attributes=None, id_prefix=None, user=None):
        """Filter entity objects from backend."""
        try:
            return self.server.backend.filter_entities(
                    categories=categories, attributes=attributes,
                    id_prefix=id_prefix, user=user)
        except Entity.DoesNotExist as e:
            raise HttpRequestError(hrc.NOT_FOUND(e))
        except ServerBackend.InvalidOperation as e:
            raise HttpRequestError(hrc.BAD_REQUEST(e))
        except ServerBackend.ServerBackendError as e:
            print e
            raise HttpRequestError(hrc.SERVER_ERROR())

    def _save_entities(self, entities, id_prefix=None, user=None):
        """Save Entity objects to backend."""
        try:
            return self.server.backend.save_entities(entities, id_prefix=id_prefix, user=user)
        except Entity.DoesNotExist as e:
            raise HttpRequestError(hrc.NOT_FOUND(e))
        except ServerBackend.InvalidOperation as e:
            raise HttpRequestError(hrc.BAD_REQUEST(e))
        except ServerBackend.ServerBackendError as e:
            print e
            raise HttpRequestError(hrc.SERVER_ERROR())

    def _delete_entities(self, entity_ids, user=None):
        """Delete Entity IDs from backend."""
        try:
            return self.server.backend.delete_entities(entity_ids, user=user)
        except Entity.DoesNotExist as e:
            raise HttpRequestError(hrc.NOT_FOUND(e))
        except ServerBackend.InvalidOperation as e:
            raise HttpRequestError(hrc.BAD_REQUEST(e))
        except ServerBackend.ServerBackendError as e:
            print e
            raise HttpRequestError(hrc.SERVER_ERROR())

    def _exec_action(self, action, entity, payload=None, user=None):
        """Instruct backend to execute Action on the given Entity."""
        try:
            return self.server.backend.exec_action(action, entity, payload=payload, user=user)
        except Entity.DoesNotExist as e:
            raise HttpRequestError(hrc.NOT_FOUND(e))
        except ServerBackend.InvalidOperation as e:
            raise HttpRequestError(hrc.BAD_REQUEST(e))
        except ServerBackend.ServerBackendError as e:
            print e
            raise HttpRequestError(hrc.SERVER_ERROR())


class EntityHandler(HandlerBase):
    """HTTP handler for existing Entity instances."""

    def get(self, request, entity_id):
        """Retrieve a resource instance."""
        try:
            parser, renderer = self._request_init(request)
            entity = self._get_entity(entity_id, user=request.user)
        except HttpRequestError as e:
            return e.response

        dao = DataObject(translator=self.translator)
        dao.load_from_entity(entity)
        renderer.render(dao)

        return HttpResponse(renderer.headers, renderer.body)

    def post(self, request, entity_id):
        """Execute an Action on a resource instance."""

        # Action query argument
        try:
            action_name = request.query_args['action'][0]
        except (KeyError, IndexError):
            return hrc.BAD_REQUEST('Missing action query parameter')

        # Get instance
        try:
            parser, renderer = self._request_init(request)
            entity = self._get_entity(entity_id, user=request.user)
        except HttpRequestError as e:
            return e.response

        # Only a single data object allowed
        if not parser.objects:
            return hrc.BAD_REQUEST('No action instance specified')
        elif len(parser.objects) > 1:
            return hrc.BAD_REQUEST('More than one action instance specified')
        dao = parser.objects[0]
        dao.translator = self.translator

        # Create Action instance
        try:
            action = dao.save_as_action(category_registry=self.server.registry)
        except DataObject.Invalid as e:
            return hrc.BAD_REQUEST(e)

        # Verify Action query param
        if action.category.term != action_name:
            return hrc.BAD_REQUEST('%s: query parameter mismatch action category' % action_name)

        # Verify Action is applicable
        if not entity.occi_is_applicable_action(action.category):
            return hrc.BAD_REQUEST('%s: action not applicable' % action_name)

        # Execute Action
        try:
            body = self._exec_action(action, entity, payload=request.body, user=request.user)
        except HttpRequestError as e:
            return e.response

        # Build response
        headers = []
        if body:
            headers.append(('Content-Type', 'application/octet-stream'))
            response = HttpResponse(headers, body)
        else:
            response = hrc.ALL_OK()
        return response

    def put(self, request, entity_id):
        """Update an existing resource instance."""
        # Parse request
        try:
            parser, renderer = self._request_init(request)
        except HttpRequestError as e:
            return e.response

        # Only a single data object allowed
        if not parser.objects:
            return hrc.BAD_REQUEST('No resource instance specified')
        elif len(parser.objects) > 1:
            return hrc.BAD_REQUEST('More than one resource instance specified')
        dao = parser.objects[0]
        dao.translator = self.translator

        # Load entity object from backend
        try:
            entity = self._get_entity(entity_id, user=request.user)
        except HttpRequestError as e:
            return e.response

        # Update entity object from request data
        try:
            dao.save_to_entity(entity=entity,
                    category_registry=self.server.registry)
        except DataObject.Invalid as e:
            return hrc.BAD_REQUEST(e)

        # Save the updated entity object
        try:
            id_list = self._save_entities([entity], user=request.user)
        except HttpRequestError as e:
            return e.response

        return hrc.ALL_OK()

    def delete(self, request, entity_id):
        """Delete a resource instance."""
        try:
            parser, renderer = self._request_init(request)
            self._delete_entities([entity_id])
        except HttpRequestError as e:
            return e.response

        return hrc.ALL_OK()

class CollectionHandler(HandlerBase):
    """HTTP handler for collections."""

    def get(self, request, path):
        """Get the collection of resource instances under the specified path."""
        # Parse request
        try:
            parser, renderer = self._request_init(request)
        except HttpRequestError as e:
            return e.response

        # Filtering parameters
        categories = []
        attributes = []
        id_prefix = None

        # Can path be mapped to a Kind/Mixin location?
        t = self.server.registry.lookup_location(path)
        if t:
            categories.append(t)
        else:
            id_prefix = path.lstrip('/')

        # Category and attribute filters
        if parser.objects:
            dao = parser.objects[0]
            dao.translator = self.translator
            for category in dao.categories:
                try:
                    categories.append(self.server.registry.lookup_id(category))
                except Category.DoesNotExist as e:
                    return hrc.BAD_REQUEST(e)
            # FIXME - what about converting value to indicated type?
            attributes = dao.attributes

        # Retrieve resource instances from backend
        try:
            entities = self._filter_entities(categories=categories, attributes=attributes,
                    id_prefix=id_prefix, user=request.user)
        except HttpRequestError as e:
            return e.response

        # Render response
        objects = []
        for entity in entities:
            dao = DataObject(translator=self.translator)
            dao.load_from_entity(entity)
            objects.append(dao)
        renderer.render(objects)

        return HttpResponse(renderer.headers, renderer.body)

    def post(self, request, path):
        """Create new resource instance(s) OR execute an action on the specified
        collection.
        """
        # Action request?
        if request.query_args:
            return self._collection_action(request, path)

        # Parse request
        try:
            parser, renderer = self._request_init(request)
        except HttpRequestError as e:
            return e.response

        # Any data-objects submitted?
        if not parser.objects:
            return hrc.BAD_REQUEST('No resource instance(s) specified')

        # Convert request objects to entity instances
        entities = []
        try:
            for dao in parser.objects:
                dao.translator = self.translator
                entities.append(dao.save_to_entity(
                    category_registry=self.server.registry))
        except DataObject.Invalid as e:
            return hrc.BAD_REQUEST(e)
        except Entity.EntityError as e:
            return hrc.BAD_REQUEST(e)

        # Save all entities using a single backend operation
        try:
            id_list = self._save_entities(entities, id_prefix=path, user=request.user)
        except HttpRequestError as e:
            return e.response

        # Response is a list of locations
        dao_list = []
        for entity_id in id_list:
            dao_list.append(DataObject(
                location=self.translator.from_native(entity_id)))

        # Render response
        renderer.render(dao_list)

        # Set Location header to the first ID
        renderer.headers.append(('Location', self.translator.from_native(id_list[0])))

        return HttpResponse(renderer.headers, renderer.body)

    def _collection_action(self, request, path):
        return hrc.NOT_IMPLEMENTED()

    def _update_mixin_collection(self, request, path, add=True):
        # Parse request
        parser, renderer = self._request_init(request)

        # Lookup Mixin at the specified path
        mixin = self.server.registry.lookup_location(path)
        if not mixin:
            raise HttpRequestError(hrc.NOT_FOUND('%s: no such collection' % path))
        if not isinstance(mixin, Mixin):
            raise HttpRequestError(hrc.BAD_REQUEST('%s: not a mixin location' % path))

        # Get entities corresponding to the given locations
        entities = []
        for dao in parser.objects:
            if not dao.location:
                return hrc.BAD_REQUEST('resource instance location expected')
            entity_id = self.translator.to_native(dao.location)
            entity = self._get_entity(entity_id, user=request.user)
            if add:
                entity.occi_add_mixin(mixin)
            else:
                entity.occi_remove_mixin(mixin)
            entities.append(entity)

        # Save all updated entities using a single backend operation
        try:
            self._save_entities(entities, user=request.user)
        except HttpRequestError as e:
            return e.response

    def put(self, request, path):
        """Add resource instance(s) to Mixin collection"""
        try:
            self._update_mixin_collection(request, path, add=True)
        except HttpRequestError as e:
            return e.response
        return hrc.ALL_OK()

    def delete(self, request, path):
        """Remove resource instance(s) from Mixin collection"""
        try:
            self._update_mixin_collection(request, path, add=False)
        except HttpRequestError as e:
            return e.response
        return hrc.ALL_OK()

class DiscoveryHandler(HandlerBase):
    """HTTP handler for the OCCI discovery interface."""

    def get(self, request):
        """List all Category instance registered in the system"""
        # Parse request
        try:
            parser, renderer = self._request_init(request)
        except HttpRequestError as e:
            return e.response

        dao = DataObject(translator=self.translator,
                categories=self.server.registry.all())
        dao.render_flags['category_discovery'] = True

        # Render response
        renderer.render(dao)
        return HttpResponse(renderer.headers, renderer.body)

    def post(self, request):
        """Http POST not valid for the discovery interface."""
        return hrc.BAD_REQUEST()

    def put(self, request):
        """Create user-defined Mixin instance(s)"""
        # Parse request and extract Mixin defs
        try:
            parser, renderer = self._request_init(request)
        except HttpRequestError as e:
            return e.response

        # Any Mixin defs supplied?
        if not parser.objects:
            return hrc.BAD_REQUEST('No Mixin definition(s) supplied')

        # Extract Mixin definitions
        mixins = []
        try:
            for category in parser.objects[0].categories:
                # Required parameters
                if not hasattr(category, 'related'):
                    return hrc.BAD_REQUEST('"%s": not a Mixin' % category)
                if not hasattr(category, 'location') or not category.location:
                    return hrc.BAD_REQUEST('"%s": location must be specified' % category)

                # Mixin must not exist
                try:
                    self.server.registry.lookup_id(category)
                except Category.DoesNotExist:
                    pass
                else:
                    return hrc.BAD_REQUEST('%s: Category already exist' % category)

                # Mixin location must not be used
                location = self.translator.to_native(category.location) 
                if self.server.registry.lookup_location(location):
                    return hrc.BAD_REQUEST('%s: conflicting location path' % location)

                # Create Mixin instance
                mixin = Mixin(category.term, category.scheme,
                        userdefined=True,
                        related=category.related, location=location)
                mixins.append(mixin)
        except Category.Invalid as e:
            return hrc.BAD_REQUEST(e)

        # Store Mixins in backend and save to Category registry
        try:
            for mixin in mixins:
                mixin = self.server.backend.add_user_mixin(mixin, user=request.user)
                try:
                    # FIXME: locking needed to avoid race condition
                    self.server.registry.register(mixin)
                except Category.Invalid as e:
                    self.server.backend.remove_user_mixin(mixin, user=request.user)
                    return hrc.BAD_REQUEST(e)
        except ServerBackend.InvalidOperation as e:
            return hrc.BAD_REQUEST(e)
        except ServerBackend.ServerBackendError as e:
            print e
            return hrc.SERVER_ERROR()
        except NotImplementedError as e:
            return hrc.NOT_IMPLEMENTED(e)

        return hrc.ALL_OK()

    def delete(self, request):
        """Remove user-defined Mixin instance(s)"""
        # Parse request and extract Mixin defs
        try:
            parser, renderer = self._request_init(request)
        except HttpRequestError as e:
            return e.response

        # Any Mixin defs supplied?
        if not parser.objects:
            return hrc.BAD_REQUEST('No Mixin definition(s) supplied')

        # Extract Mixin definitions
        mixins = []
        try:
            for category in parser.objects[0].categories:
                # Find Mixin in registry
                try:
                    mixin = self.server.registry.lookup_id(category)
                except Category.DoesNotExist:
                    return hrc.BAD_REQUEST('%s: does not exist' % category)

                if not hasattr(mixin, 'userdefined') or not mixin.userdefined:
                    return hrc.BAD_REQUEST('%s: not a user-defined Mixin' % mixin)
                mixins.append(mixin)
        except Category.Invalid as e:
            return hrc.BAD_REQUEST(e)

        # Remove Mixins from backend and Category registry
        try:
            for mixin in mixins:
                self.server.backend.remove_user_mixin(mixin, user=request.user)
                try:
                    self.server.registry.unregister(mixin)
                except Category.Invalid as e:
                    return hrc.BAD_REQUEST(e)
        except ServerBackend.InvalidOperation as e:
            return hrc.BAD_REQUEST(e)
        except ServerBackend.ServerBackendError as e:
            print e
            return hrc.SERVER_ERROR()
        except NotImplementedError as e:
            return hrc.NOT_IMPLEMENTED(e)

        return hrc.ALL_OK()



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

import uuid
from abc import ABCMeta
try:
    from ordereddict import OrderedDict
except ImportError:
    OrderedDict = dict

from occi.core import Entity, Resource, Link, CategoryRegistry

class OCCIServer(object):
    """An OCCI Server instance."""

    def __init__(self, backend):
        """A `ServerBackend` object is required to instantiate an OCCI Server."""
        self.registry = CategoryRegistry()
        self.backend = backend

class ServerBackend(object):
    __metaclass__ = ABCMeta

    class ServerBackendError(Exception):
        pass
    class InvalidOperation(ServerBackendError):
        pass

    def get_entity(self, entitiy_id, user=None):
        raise self.ServerBackendError('Server Backend must implement get_entity')

    def filter_entities(self, categories=None, attributes=None, id_prefix=None, user=None):
        """Return a list of `Entity` objects matching the specified filter.

        The filter parameters are specified using the keyword arguments
        described below.  All specified filter parameters must match the
        returned `Entity` instances.

        :keyword categories: A list of `Category` instances a matching `Entity`
            instance must be a member of.
        :keyword attributes: A list of attribute key-value pairs which must all
            be present in a matching `Entity` instance.
        :keyword id_prefix: A string prefix which must match the start of the
            `Entity` ID.
        :keyword user: The authenticated user.
        :return: A list of `Entity` instances matching the filter parameters.
        """
        raise self.ServerBackendError('Server Backend must implement filter_entities')

    def save_entities(self, entities, id_prefix=None, user=None):
        """Save a set of entities (resource instances) in a single atomic
        operation.

        :param entities: A list of `Entity` objects to persist.
        :keyword id_prefix: `Entity` ID prefix suggested by client for new object.
        :keyword user: The authenticated user.
        :return: A list IDs of the saved `Entity` objects.
        """
        raise self.ServerBackendError('Server Backend must implement save_entities')

    def delete_entities(self, entity_ids, user=None):
        """Delete a set of entities (resource instances) in a single atomic
        operation.

        :param entity_ids: A list `Entity` IDs to delete.
        :keyword user: The authenticated user.
        """
        raise self.ServerBackendError('Server Backend must implement delete_entities')

    def exec_action(self, action, entity, payload=None, user=None):
        """Execute `Action` on the given `Entity` (resource instance).

        :param action: `Action` instance.
        :param entity: `Entity` (resource) instance.
        :keyword payload: Binary payload supplied with Action.
        :keyword user: The authenticated user.
        """
        raise self.ServerBackendError('Server Backend must implement exec_action')

    def exec_action_on_collection(self, action, collection, payload=None, user=None):
        """Execute `Action` on the all `Entity` instances in the specified
        collection (if applicable).

        :param action: `Action` instance.
        :param collection: `Kind` or `Mixin` instance.
        :keyword payload: Binary payload supplied with Action.
        :keyword user: The authenticated user.
        """
        raise self.ServerBackendError('Server Backend must implement exec_action_on_collection')

    def add_user_mixin(self, mixin, user=None):
        """Validate a user-supplied Mixin instance and perform any
        backend-specific tasks related to the event. The method must return the
        mixin instance to be added, either a modified version or the original
        user-supplied Mixin instance.

        :param mixin: User-defined `Mixin` instance.
        :keyword user: The authenticated user.
        :return: The actual `Mixin` instance to be added.
        """
        raise NotImplementedError('User-defined Mixins not supported')

    def remove_user_mixin(self, mixin, user=None):
        """Validate the removal of a user-defined `Mixin` instance and perform
        any backend-specific tasks related to the event. The method is expected
        to throw a InvalidOperation exception if the removal operation is to be
        refused.

        :param mixin: User-defined `Mixin` instance to be removed.
        :keyword user: The authenticated user.
        """
        raise NotImplementedError('User-defined Mixins not supported')


class DummyBackend(ServerBackend):
    """Very simple (and inefficient) in-memory backend for test purposes.

    >>> backend = DummyBackend()
    >>> from occi.ext.infrastructure import *
    >>> t = backend.save_entities([ComputeKind.entity_type(ComputeKind)])
    >>> compute = ComputeKind.entity_type(ComputeKind)
    >>> compute.occi_set_attributes([('occi.compute.memory', '2.0')])
    >>> storage = StorageKind.entity_type(StorageKind)
    >>> compute_id, storage_id = backend.save_entities([compute, storage])
    >>> link = StorageLinkKind.entity_type(StorageLinkKind)
    >>> link.occi_set_attributes([('source', compute_id), ('target', storage_id), ('occi.storagelink.deviceid', 'ide:0:0')])
    >>> link_id = backend.save_entities([link])
    >>> len(backend.filter_entities())
    4
    >>> len(backend.filter_entities(categories=[ComputeKind]))
    2
    >>> len(backend.filter_entities(categories=[ComputeKind], attributes=[('occi.compute.memory', 2.0)]))
    1
    >>> backend.get_entity(compute_id) == compute
    True
    >>> backend.delete_entities(t)
    >>> [entity.id for entity in backend.filter_entities(categories=[ComputeKind])] == [compute_id]
    True
    """

    def __init__(self):
        self._db = OrderedDict()

    def get_entity(self, entity_id, user=None):
        try:
            return self._db[entity_id]
        except KeyError:
            raise Entity.DoesNotExist(entity_id)

    def filter_entities(self, categories=None, attributes=None, id_prefix=None, user=None):
        result = []
        for entity_id, entity in self._db.iteritems():
            skip = False
            # Filter on id_prefix
            if id_prefix:
                if not entity_id.startswith(id_prefix):
                    continue

            # Filter on Categories
            cats = entity.occi_list_categories()
            for cat in categories or ():
                if str(cat) not in cats:
                    skip=True
                    break
            if skip: continue

            # Filter on Attributes
            if categories and attributes:
                for name, value in attributes:
                    t = entity.occi_get_attribute(name)
                    if str(t) != str(value):    # FIXME - this implies "2.0" == 2.0
                        skip = True
                        break
            if skip: continue

            result.append(entity)

        return result

    def save_entities(self, entities, id_prefix=None, user=None):
        id_list = []
        for entity in entities:
            # Generate ID if new instance
            if not entity.id:
                loc = entity.occi_get_kind().location or ''
                entity.id = '%s%s' % (loc, uuid.uuid4())

            # Links
            if isinstance(entity, Link):
                source = self.get_entity(entity.occi_get_attribute('source'), user=user)
                target = self.get_entity(entity.occi_get_attribute('target'), user=user)
                entity.source = source
                entity.target = target
                links = []
                for l in source.links:
                    if l.id != source.id:
                        links.append(l)
                links.append(entity)
                source.links = links

            self._db[entity.id] = entity
            id_list.append(entity.id)
        return id_list

    def delete_entities(self, entity_ids, user=None):
        for entity_id in entity_ids:
            try:
                entity = self._db[entity_id]
                if isinstance(entity, Resource):
                    for l in entity.links:
                        self._db.pop(l.id, None)
                elif isinstance(entity, Link):
                    try:
                        entity.source.links.remove(entity)
                    except ValueError:
                        pass
                del self._db[entity_id]
            except KeyError:
                raise Entity.DoesNotExist(entity_id)

    def exec_action(self, action, entity, payload=None, user=None):
        try:
            return getattr(entity, 'exec_action')(action, payload=payload)
        except AttributeError:
            return None

    def add_user_mixin(self, mixin, user=None):
        return mixin

    def remove_user_mixin(self, mixin, user=None):
        pass

if __name__ == "__main__":
    import doctest
    doctest.testmod()



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

from occi.core import (Category, ExtCategory, Kind, Mixin, Resource, Link,
        ResourceKind, LinkKind, Attribute, IntAttribute, FloatAttribute)

#
# Compute Kind
# ============

ComputeStartActionCategory = Category('start', 'http://schemas.ogf.org/occi/infrastructure/compute/action#',
        title='Start Compute Resource',
        attributes=(
            Attribute('method', required=False, mutable=True),
        ))
ComputeStopActionCategory = Category('stop', 'http://schemas.ogf.org/occi/infrastructure/compute/action#',
        title='Stop Compute Resource',
        attributes=(
            Attribute('method', required=False, mutable=True),
        ))
ComputeRestartActionCategory = Category('restart', 'http://schemas.ogf.org/occi/infrastructure/compute/action#',
        title='Restart Compute Resource',
        attributes=(
            Attribute('method', required=False, mutable=True),
        ))
ComputeSuspendActionCategory = Category('suspend', 'http://schemas.ogf.org/occi/infrastructure/compute/action#',
        title='Suspend Compute Resource',
        attributes=(
            Attribute('method', required=False, mutable=True),
        ))

ComputeKind = Kind('compute', 'http://schemas.ogf.org/occi/infrastructure#',
        title='Compute Resource',
        related=ResourceKind,
        entity_type=Resource,
        location='compute/',
        attributes=(
            Attribute('occi.compute.architecture', required=False, mutable=False),
            IntAttribute('occi.compute.cores', required=False, mutable=True),
            Attribute('occi.compute.hostname', required=False, mutable=True),
            FloatAttribute('occi.compute.speed', required=False, mutable=True),
            FloatAttribute('occi.compute.memory', required=False, mutable=True),
            Attribute('occi.compute.state', required=False, mutable=False),
        ),
        actions=(
            ComputeStartActionCategory,
            ComputeStopActionCategory,
            ComputeRestartActionCategory,
            ComputeSuspendActionCategory,
        ),
)

#
# Network Kind
# ============

NetworkUpActionCategory = Category('up', 'http://schemas.ogf.org/occi/infrastructure/network/action#',
        title='Bring up Network Resource')
NetworkDownActionCategory = Category('down', 'http://schemas.ogf.org/occi/infrastructure/network/action#',
        title='Take down Network Resource')

NetworkKind = Kind('network', 'http://schemas.ogf.org/occi/infrastructure#',
        title='Network Resource',
        related=ResourceKind,
        entity_type=Resource,
        location='network/',
        attributes=(
            IntAttribute('occi.network.vlan', required=False, mutable=True),
            Attribute('occi.network.label', required=False, mutable=True),
            Attribute('occi.network.state', required=False, mutable=False),
        ),
        actions=(
            NetworkUpActionCategory,
            NetworkDownActionCategory,
        ),
)

#
# IPNetwork Mixin
# ===============

IPNetworkMixin = Mixin('ipnetwork', 'http://schemas.ogf.org/occi/infrastructure/network#',
        title='IPNetworking Mixin',
        location='ipnetwork/',
        attributes=(
            Attribute('occi.network.address', required=False, mutable=True),
            Attribute('occi.network.gateway', required=False, mutable=True),
            Attribute('occi.network.allocation', required=False, mutable=True),
        )
)

#
# Storage Kind
# ============

StorageOnlineActionCategory = Category('online',
        'http://schemas.ogf.org/occi/infrastructure/storage/action#',
        title='Bring Storage Resource online')
StorageOfflineActionCategory = Category('offline',
        'http://schemas.ogf.org/occi/infrastructure/storage/action#',
        title='Bring Storage Resource offline')
StorageBackupActionCategory = Category('backup',
        'http://schemas.ogf.org/occi/infrastructure/storage/action#',
        title='Backup Storage Resource')
StorageSnapshotActionCategory = Category('snapshot',
        'http://schemas.ogf.org/occi/infrastructure/storage/action#',
        title='Take snapshot of Storage Resource')
StorageResizeActionCategory = Category('resize',
        'http://schemas.ogf.org/occi/infrastructure/storage/action#',
        title='Resize Storage Resource',
        attributes=(
            Attribute('size', required=True, mutable=True),
        ))

StorageKind = Kind('storage', 'http://schemas.ogf.org/occi/infrastructure#',
        title='Storage Resource',
        related=ResourceKind,
        entity_type=Resource,
        location='storage/',
        attributes=(
            FloatAttribute('occi.storage.size', required=True, mutable=True),
            Attribute('occi.storage.state', required=False, mutable=False),
        ),
        actions=(
            StorageOnlineActionCategory,
            StorageOfflineActionCategory,
            StorageBackupActionCategory,
            StorageSnapshotActionCategory,
            StorageResizeActionCategory,
        ),
)

#
# NetworkInterface Kind
# =====================

NetworkInterfaceKind = Kind('networkinterface', 'http://schemas.ogf.org/occi/infrastructure#',
        title='NetworkInterface Link',
        related=LinkKind,
        entity_type=Link,
        location='link/networkinterface/',
        attributes=(
            Attribute('occi.networkinterface.interface', required=False, mutable=True),
            Attribute('occi.networkinterface.mac', required=False, mutable=True),
            Attribute('occi.networkinterface.state', required=False, mutable=False),
        )
)

#
# IPNetworkInterface Mixin
# ========================

IPNetworkInterfaceMixin = Mixin('ipnetworkinterface', 'http://schemas.ogf.org/occi/infrastructure#',
        title='IPNetworkInterface Link',
        location='link/ipnetworkinterface/',
        attributes=(
            Attribute('occi.networkinterface.ip', required=False, mutable=True),
            Attribute('occi.networkinterface.gateway', required=False, mutable=True),
            Attribute('occi.networkinterface.allocation', required=True, mutable=True),
        )
)

#
# Storage Kind
# ============

StorageLinkKind = Kind('storagelink', 'http://schemas.ogf.org/occi/infrastructure#',
        title='Storage Link',
        related=LinkKind,
        entity_type=Link,
        location='link/storage/',
        attributes=(
            Attribute('occi.storagelink.deviceid', required=True, mutable=True),
            Attribute('occi.storagelink.mountpoint', required=False, mutable=True),
            Attribute('occi.storagelink.state', required=False, mutable=False),
        )
)



#!/usr/bin/env python
#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

from distutils.core import setup
from occi import __version__

setup(
    name='occi',
    version=__version__,
    description='Open Cloud Computing Interface Client/Server Library',
    long_description=open('README.rst').read(),
    url='http://github.com/nyren/occi-py',
    author='Ralf Nyren',
    author_email='ralf@nyren.net',
    maintainer='Ralf Nyren',
    maintainer_email='ralf@nyren.net',
    keywords=['OCCI', 'Cloud', 'Library', 'REST', 'RESTful', 'IaaS', 'PaaS', 'SaaS'],
    license='GNU Lesser General Public License (LGPL) version 3',
    platforms=['any'],
    packages=['occi', 'occi.ext', 'occi.http'],
    scripts=['occi/demo/occi_iaas_demo.py'],
#   test_suite='tests.all',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules'],
)



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

import re

from occi.core import Category, Kind, Mixin
from occi.http.header import HttpHeaderError, HttpHeadersBase, HttpWebHeadersBase, HttpCategoryHeaders, HttpLinkHeaders, HttpAttributeHeaders
from occi.http.dataobject import DataObject, LinkRepr

_renderers= {}

class RendererError(Exception):
    pass

def register_renderer(content_type, renderer):
    _renderers[content_type] = renderer

def unregister_renderer(content_type):
    del _renderers[content_type]

def get_renderer(accept_types=None):
    """Return a renderer matching the list of accepted content-types.

    :keyword accept_types: List of acceptable content types, i.e. the result of
        _parsing_ an Accept header.

    >>> p = get_renderer(['text/occi'])
    >>> isinstance(p, HeaderRenderer)
    True
    >>> p = get_renderer(['text/plain'])
    >>> isinstance(p, TextPlainRenderer)
    True
    >>> p = get_renderer()
    >>> isinstance(p, TextRenderer)
    True
    >>> p = get_renderer(['text/html', '*/*'])
    >>> isinstance(p, TextRenderer)
    True
    >>> p = get_renderer(['text/html', 'image/jpeg', 'image/png'])
    Traceback (most recent call last):
        File "renderer.py", line 41, in renderer
    RendererError: No renderer found for requested content types
    """
    r = None
    if not accept_types:
        r = _renderers.get(None)
    else:
        for content_type in accept_types:
            try:
                r = _renderers[content_type]
            except KeyError:
                pass
            else:
                break

    if not r:
        raise RendererError('No renderer found for requested content types')
    return r()

class Renderer(object):
    """Renderer base class.

    A Renderer must implement the render() method which renders the given
    `DataObject` or list of `DataObject`s.

    The result of the render() method is stored in the following attributes:
    :var headers: A list of HTTP Header name-value tuples
    :var body: The HTTP Body as a string

    """
    def __init__(self):
        self.headers = []
        self.body = ''

    def render(self, objects):
        """The render method doing the actual work.

        :keyword objects: A `DataObject` or a list of `DataObject` instances
        """
        raise NotImplementedError('%s: does not implement the render() method',
                self.__class__.__name__)

class HeaderRenderer(Renderer):
    """Renderer for the text/occi content type.

    >>> from occi.ext.infrastructure import ComputeKind, StorageKind
    >>> cats = [ComputeKind]
    >>> links = [LinkRepr(target_location='http://example.com/storage/345', target_categories=[StorageKind])]
    >>> attrs = [('occi.compute.memory', '2.0'), ('occi.compute.speed', '2.667')]
    >>> obj = DataObject(location='http://example.com/compute/123', categories=cats, links=links, attributes=attrs)
    >>> r = HeaderRenderer()
    >>> r.render([obj])
    >>> r.headers
    [('Content-Type', 'text/occi'), ('X-OCCI-Location', 'http://example.com/compute/123')]
    >>> r.body
    ''
    >>> r = HeaderRenderer()
    >>> r.render(obj)
    >>> r.headers
    [('Content-Type', 'text/occi'), ('Category', 'compute; scheme="http://schemas.ogf.org/occi/infrastructure#"; class="kind"; title="Compute Resource"'), ('Link', '<http://example.com/storage/345>; rel="http://schemas.ogf.org/occi/infrastructure#storage"; title=""'), ('X-OCCI-Attribute', 'occi.compute.memory="2.0"'), ('X-OCCI-Attribute', 'occi.compute.speed="2.667"')]

    """
    def render(self, objects):
        self.headers.append(('Content-Type', 'text/occi'))
        if isinstance(objects, list) or isinstance(objects, tuple):
            self._render_obj_list(objects)
        else:
            self._render_single_obj(objects)

    def _render_single_obj(self, obj):
        """Render a single `DataObject`.
        """
        # Category headers
        category_headers = HttpCategoryHeaders()
        for category in obj.categories:
            params = []
            params.append(('scheme',  category.scheme))

            cat_class = category.__class__.__name__.lower()
            # FIXME: this is a bug in the spec, fix it?
            if cat_class == 'category': cat_class = 'action'
            params.append(('class',  cat_class))

            if category.title:
                params.append(('title',  category.title))

            if 'category_discovery' in obj.render_flags:
                if category.related:
                    params.append(('rel',  category.related))
                if category.attributes:
                    params.append(('attributes',
                        ' '.join([attr.name for attr in category.unique_attributes])))
                if hasattr(category, 'location') and category.location:
                    params.append(('location',
                        obj.translator.from_native(category.location, path_only=True)))

            category_headers.add(category.term, params)
        [self.headers.append(('Category', h)) for h in category_headers.headers()]

        # Link headers
        link_headers = HttpLinkHeaders()
        for link in obj.links:
            params = []
            params.append(('rel',  ' '.join([str(cat) for cat in link.target_categories])))
            params.append(('title',  link.target_title or ''))
            if link.link_location:
                params.append(('self',  link.link_location))
                for attr, value in link.link_attributes:
                    params.append((attr,  value))

            link_headers.add(link.target_location, params)
        [self.headers.append(('Link', h)) for h in link_headers.headers()]

        # Attribute headers
        attribute_headers = HttpAttributeHeaders()
        [attribute_headers.add(attr, value) for attr, value in obj.attributes]
        [self.headers.append(('X-OCCI-Attribute', h)) for h in attribute_headers.headers()]

    def _render_obj_list(self, objects):
        """Render a list of `DataObject` instances. Only the location of each
        `DataObject` will be displayed by the Http Header rendering.
        """
        for obj in objects:
            if not obj.location:
                raise RendererError('DataObject has no location')
            self.headers.append(('X-OCCI-Location', obj.location))

class TextPlainRenderer(HeaderRenderer):
    """Renderer for the text/plain content type.

    >>> from occi.ext.infrastructure import ComputeKind, StorageKind
    >>> cats = [ComputeKind]
    >>> links = [LinkRepr(target_location='http://example.com/storage/345', target_categories=[StorageKind])]
    >>> attrs = [('occi.compute.memory', '2.0'), ('occi.compute.speed', '2.667')]
    >>> obj = DataObject(location='http://example.com/compute/123', categories=cats, links=links, attributes=attrs)
    >>> r = TextPlainRenderer()
    >>> r.render([obj])
    >>> r.headers
    [('Content-Type', 'text/plain')]
    >>> r.body
    'X-OCCI-Location: http://example.com/compute/123\\r\\n'
    >>> r = TextPlainRenderer()
    >>> r.render(obj)
    >>> r.headers
    [('Content-Type', 'text/plain')]
    >>> r.body
    'Category: compute; scheme="http://schemas.ogf.org/occi/infrastructure#"; class="kind"; title="Compute Resource"\\r\\nLink: <http://example.com/storage/345>; rel="http://schemas.ogf.org/occi/infrastructure#storage"; title=""\\r\\nX-OCCI-Attribute: occi.compute.memory="2.0"\\r\\nX-OCCI-Attribute: occi.compute.speed="2.667"\\r\\n'

    """
    def render(self, objects):
        super(TextPlainRenderer, self).render(objects)
        for name, value in self.headers[1:]:
            self.body += '%s: %s\r\n' % (name, value)
        self.headers = []
        self.headers.append(('Content-Type', 'text/plain'))

class TextURIListRenderer(Renderer):
    """Renderer for the text/uri-list content type.

    This renderer always returns URIs, even if just one object is rendered.

    >>> from occi.ext.infrastructure import ComputeKind, StorageKind
    >>> objs = []
    >>> objs.append(DataObject(location='/compute/123', categories=[ComputeKind]))
    >>> objs.append(DataObject(location='/compute/234', categories=[ComputeKind]))
    >>> objs.append(DataObject(location='/storage/345', categories=[StorageKind]))
    >>> r = TextURIListRenderer()
    >>> r.render(objs)
    >>> r.headers
    [('Content-Type', 'text/uri-list')]
    >>> r.body
    '/compute/123\\r\\n/compute/234\\r\\n/storage/345\\r\\n'
    >>> r = TextURIListRenderer()
    >>> r.render(objs[1])
    >>> r.headers
    [('Content-Type', 'text/uri-list')]
    >>> r.body
    '/compute/234\\r\\n'
    """
    def render(self, objects):
        self.headers.append(('Content-Type', 'text/uri-list'))
        if not isinstance(objects, list) and not isinstance(objects, tuple):
            objects = [objects]
        for obj in objects:
            if obj.location:
                self.body += '%s\r\n' % obj.location

class TextRenderer(Renderer):
    """The default renderer. Uses text/plain for single object rendering and
    text/uri-list for multiple objects rendering.

    >>> from occi.ext.infrastructure import ComputeKind, StorageKind
    >>> objs = []
    >>> attrs = [('occi.compute.memory', '2.0'), ('occi.compute.speed', '2.667')]
    >>> objs.append(DataObject(location='/compute/123', categories=[ComputeKind]))
    >>> objs.append(DataObject(location='/compute/234', categories=[ComputeKind], attributes=attrs))
    >>> objs.append(DataObject(location='/storage/345', categories=[StorageKind]))
    >>> r = TextRenderer()
    >>> r.render(objs)
    >>> r.headers
    [('Content-Type', 'text/uri-list')]
    >>> r.body
    '/compute/123\\r\\n/compute/234\\r\\n/storage/345\\r\\n'
    >>> r = TextRenderer()
    >>> r.render(objs[1])
    >>> r.headers
    [('Content-Type', 'text/plain')]
    >>> r.body
    'Category: compute; scheme="http://schemas.ogf.org/occi/infrastructure#"; class="kind"; title="Compute Resource"\\r\\nX-OCCI-Attribute: occi.compute.memory="2.0"\\r\\nX-OCCI-Attribute: occi.compute.speed="2.667"\\r\\n'
    """
    def render(self, objects):
        if isinstance(objects, list) or isinstance(objects, tuple):
            r = TextURIListRenderer()
        else:
            r = TextPlainRenderer()
        r.render(objects)
        self.headers = r.headers
        self.body = r.body

# Register required renderers
register_renderer(None, TextRenderer)
register_renderer('text/occi', HeaderRenderer)
register_renderer('text/plain', TextPlainRenderer)
register_renderer('text/uri-list', TextURIListRenderer)
register_renderer('text/*', TextRenderer)
register_renderer('*/*', TextRenderer)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#
# Copyright (C) 2010-2011  Ralf Nyren <ralf@nyren.net>
#
# This file is part of the occi-py library.
#
# The occi-py library is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The occi-py library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with the occi-py library.  If not, see <http://www.gnu.org/licenses/>.
#

import re

from occi.core import Category, Kind, Mixin
from occi.server import OCCIServer
from occi.http.header import (HttpHeaderError, HttpHeadersBase,
        HttpWebHeadersBase, HttpCategoryHeaders, HttpLinkHeaders,
        HttpAttributeHeaders, HttpAcceptHeaders)
from occi.http.dataobject import DataObject, LinkRepr, URLTranslator

_parsers = {}

class ParserError(Exception):
    pass

def register_parser(content_type, parser):
    _parsers[content_type] = parser

def unregister_parser(content_type):
    del _parsers[content_type]

def get_parser(content_type=None, translator=None):
    """Return a parser for the given Content-Type.

    >>> p = get_parser('text/occi')
    >>> isinstance(p, HeaderParser)
    True
    >>> p = get_parser('text/plain; charset=utf-8')
    >>> isinstance(p, TextPlainParser)
    True
    >>> p = get_parser()
    >>> isinstance(p, HeaderParser)
    True
    >>> p = get_parser('application/not-supported')
    Traceback (most recent call last):
        File "parser.py", line 41, in parser
    ParserError: "application/not-supported": Content-Type not supported
    """
    p = None
    if not content_type:
        p = _parsers.get(None)
    else:
        h = HttpWebHeadersBase()
        h.parse(content_type or '')
        for c_type, c_params in h.all():
            try:
                p = _parsers[c_type]
            except KeyError:
                pass
            else:
                break

    if not p:
        raise ParserError('"%s": Content-Type not supported' % content_type)
    return p(translator=translator)

class Parser(object):
    """Parser base class.

    A Parser must implement the parse() method which parses the provided HTTP
    Headers and/or HTTP body into a set of data objects and/or location URLs.

    The result of the parse() method is stored in the following attributes:
    :var objects: A list of `DataObject` instances
    :var accept_types: A list of content types (populated by Parser.parse())

    """
    def __init__(self, translator=None):
        self.objects = []
        self.accept_types = []
        self.translator = translator or URLTranslator('')

    def parse(self, headers=None, body=None):
        """The parse method doing the actual work. This method must be called
        from the child-class parse method.

        :keyword headers: HTTP Headers represented as a list of tuples
            containing the header name and value of each header line.
        :keyword body: HTTP Body as a string.
        """
        # Extract and parse Accept header
        for name, value in headers or ():
            if name.lower() == 'accept':
                self._parse_accept_header(value)

    def _parse_accept_header(self, header_value):
        """Parse Accept header and store the accepted content types in
        :var accept_types:
        """
        h = HttpAcceptHeaders()
        h.parse(header_value)
        for c_type, c_params in h.all_sorted():
            self.accept_types.append(c_type)

class HeaderParser(Parser):
    """Parser for the text/occi content type.

    Data is transmitted in the HTTP Header. Only a single data object can be
    represented using the text/occi content type.

    >>> headers = []
    >>> headers.append(('Category', 'network; scheme="http://schemes.ogf.org/occi/infrastructure#"; class="kind", ipnetwork; scheme="http://schemes.ogf.org/occi/infrastructure#";'))
    >>> headers.append(('Category', 'user; scheme="http://schemes.ogf.org/occi/credentials#"; class="mixin"'))
    >>> headers.append(('X-OCCI-Attribute', 'title="Test Network, testing", occi.network.label=intranet'))
    >>> headers.append(('X-OCCI-Attribute', 'occi.network.address="192.168.1.123", occi.netwark.gateway="192.168.1.1"'))
    >>> p = HeaderParser()
    >>> p.parse(headers=headers)
    >>> p.objects[0].categories
    [Kind('network', 'http://schemes.ogf.org/occi/infrastructure#'), Category('ipnetwork', 'http://schemes.ogf.org/occi/infrastructure#'), Mixin('user', 'http://schemes.ogf.org/occi/credentials#')]
    >>> p.objects[0].links
    []
    >>> p.objects[0].attributes
    [('title', 'Test Network, testing'), ('occi.network.label', 'intranet'), ('occi.network.address', '192.168.1.123'), ('occi.netwark.gateway', '192.168.1.1')]

    >>> p.objects[0].location

    """
    def parse(self, headers=None, body=None):
        categories = []
        links = []
        attributes = []
        locations = []

        # Walk list of HTTP header name-value pairs
        for name, value in headers or ():
            name = name.lower()

            if name == 'accept':
                self._parse_accept_header(value)
            elif name == 'category':
                categories.extend(self._parse_category_header(value))
            elif name == 'link':
                # FIXME - not allowing Link create/update using POST/PUT yet
                pass
            elif name == 'x-occi-attribute':
                attribute_headers = HttpAttributeHeaders()
                attribute_headers.parse(value)
                attributes.extend(attribute_headers.all())
            elif name == 'x-occi-location':
                location_headers = HttpHeadersBase()
                location_headers.parse(value)
                locations.extend(location_headers.all())

        # Only possible to represent one "full" data object with HTTP Headers.
        # Multiple data objects can only be represented with a location.
        locations = locations or [None]
        self.objects.append(DataObject(
            categories=categories,
            links=links,
            attributes=attributes,
            location=locations[0]))
        for loc in locations[1:]:
            self.objects.append(DataObject(location=loc))

    def _parse_category_header(self, header_value):
        categories = []
        category_headers = HttpCategoryHeaders()
        category_headers.parse(header_value)
        for term, attributes in category_headers.all():
            param = {}
            for k, v in attributes:
                param[k] = v

            # Category scheme
            try:
                scheme = param['scheme']
            except KeyError:
                raise HttpHeaderError('Category scheme not specified')

            # Category, Kind or Mixin?
            t = param.get('class')
            cls = Category
            if t and t.lower() == 'kind':
                cls = Kind
            elif t and t.lower() == 'mixin':
                cls = Mixin
            elif not t and param.get('location'):
                cls = Mixin

            # Related Kind/Mixin
            try:
                t = param['rel']
                r_scheme, r_term = t.split('#', 1)
                r_scheme += '#'
                related = cls(r_term, r_scheme)
            except KeyError, ValueError:
                related = None

            # Supported attributes (mutable)
            try:
                attributes = param['attributes'].split()
            except KeyError:
                attributes = None

            # Supported actions
            try:
                t =  param['actions'].split()
            except KeyError:
                actions = None
            else:
                actions = []
                for action in param['actions'].split():
                    try:
                        a_term, a_scheme = action.split('#', 1)
                    except ValueError:
                        pass
                    else:
                        actions.append(Category(a_term, a_scheme))

            # Keyword args for Category
            kwargs = {
                    'title': param.get('title'),
                    'attributes': attributes
            }

            # Additional keyword args for Kind/Mixin
            if cls == Kind or cls == Mixin:
                kwargs['related'] = related
                kwargs['actions'] = actions
                location = param.get('location')
                if location:
                    kwargs['location'] = self.translator.to_native(location)

            # Append instance to categories list
            try:
                categories.append(cls(term, scheme, **kwargs))
            except Category.Invalid as e:
                raise HttpHeaderError('%s: Invalid Category header: %s' % (scheme+term, e))

        return categories

class TextPlainParser(Parser):
    """Parser for the text/plain content type.

    Data is transmitted in the HTTP Body.

    >>> headers = [('Accept', 'text/*, */*;q=0.1')]
    >>> body  = 'Category: network; scheme="http://schemes.ogf.org/occi/infrastructure#";\\r\\n'
    >>> body += '    class=kind;\\r\\n'
    >>> body += '    title="Network Resource";\\r\\n'
    >>> body += '    rel="http://schemes.ogf.org/occi/core#resource";\\r\\n'
    >>> body += '    attributes="occi.network.vlan occi.network.label";\\r\\n'
    >>> body += '    actions="http://schemas.ogf.org/occi/infrastructure/network/action#up http://schemas.ogf.org/occi/infrastructure/network/action#down"\\n'
    >>> body += 'Category: ipnetwork; scheme="http://schemes.ogf.org/occi/infrastructure#"; class="mixin"\\r\\n'
    >>> body += 'X-OCCI-Attribute: title="Test Network, testing"\\n'
    >>> body += 'X-OCCI-Attribute: occi.network.label=intranet\\r\\n'
    >>> body += 'X-OCCI-Attribute: occi.network.address="192.168.1.123", occi.netwark.gateway="192.168.1.1"\\r\\n'
    >>> p = TextPlainParser()
    >>> p.parse(headers=headers, body=body)
    >>> p.accept_types
    ['*/*', 'text/*']
    >>> p.objects[0].categories
    [Kind('network', 'http://schemes.ogf.org/occi/infrastructure#'), Mixin('ipnetwork', 'http://schemes.ogf.org/occi/infrastructure#')]
    >>> p.objects[0].links
    []
    >>> p.objects[0].attributes
    [('title', 'Test Network, testing'), ('occi.network.label', 'intranet'), ('occi.network.address', '192.168.1.123'), ('occi.netwark.gateway', '192.168.1.1')]
    >>> body  = 'X-OCCI-Location: http://example.com/network/123\\n'
    >>> body += 'X-OCCI-Location: http://example.com/network/234\\n'
    >>> body += 'X-OCCI-Location: http://example.com/network/345\\n'
    >>> p = TextPlainParser()
    >>> p.parse(body=body)
    >>> [obj.location for obj in p.objects]
    ['http://example.com/network/123', 'http://example.com/network/234', 'http://example.com/network/345']
    """
    def __init__(self, translator=None):
        self._header_parser = HeaderParser(translator=translator)

    def get_objects(self):
        return self._header_parser.objects
    objects = property(get_objects)

    def get_accept_types(self):
        return self._header_parser.accept_types
    accept_types = property(get_accept_types)

    def parse(self, headers=None, body=None):
        super(TextPlainParser, self).parse(headers, body)
        body = body or ''
        headers = []
        for h in re.sub(r'\n\s', ' ', body.replace('\r', '')).split('\n'):
            if not h.strip():
                continue
            try:
                name, value = h.split(':', 1)
            except ValueError:
                raise HttpHeaderError(h)
            if not value:
                raise HttpHeaderError(h)

            headers.append((name, value))

        self._header_parser.parse(headers=headers)

class TextURIListParser(Parser):
    """Parser for the text/uri-list content type.

    Data is transmitted in the HTTP Body.

    >>> body  = 'http://example.com/network/123\\n'
    >>> body += 'http://example.com/network/234\\r\\n'
    >>> body += 'http://example.com/network/345 \\n'
    >>> body += '\\r\\n'
    >>> p = TextURIListParser()
    >>> p.parse(body=body)
    >>> [obj.location for obj in p.objects]
    ['http://example.com/network/123', 'http://example.com/network/234', 'http://example.com/network/345']
    """
    def parse(self, headers=None, body=None):
        super(TextURIListParser, self).parse(headers, body)
        body = body or ''
        for loc in body.replace('\r', '').split('\n'):
            loc = loc.strip()
            if loc:
                self.objects.append(DataObject(location=loc))

# Register required parsers
register_parser(None, HeaderParser)
register_parser('text/occi', HeaderParser)
register_parser('text/plain', TextPlainParser)
register_parser('text/uri-list', TextURIListParser)


if __name__ == "__main__":
    import doctest
    doctest.testmod()
